/*
 * Pocket CGI library
 * (c) 2001 TAKANO Ryousei
 * $Id: cgi.c,v 1.2 2001/07/31 08:50:01 takano Exp $
 * This library is based on the followed CGI library.
 * see: http://www.fuel.co.jp/libman.html
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "cgi.h"

// read query string 
char* get_qs()
{
  char* qs;
  char tc;
  char* p;

  p = getenv("REQUEST_METHOD");
  if (!p) {
    return NULL;
  }

  if (!strcmp(p, "GET")) {
    p = getenv("QUERY_STRING");
    qs = strdup(p);
  } else {
    char* slen = getenv("CONTENT_LENGTH");
    int len = atol(slen);
    qs = (char*)malloc(sizeof(char) * (len+1));
    fgets(qs, len+1, stdin);
  }
  return qs;
}

char* encode_qs(const char* q)
{
  const char tbl[] = "0123456789ABCDEF";
  char *buf, *p;
  p = buf = (char*)malloc(sizeof(char) * strlen(q) * 3); // ...
  while (*q) {
    if (isalnum(*q)) {
      *p++ = *q;
    } else if (*q == ' ') {
      *p++ = '+';
    } else {
      *p++ = '%';
      *p++ = tbl[(unsigned char)(*q) >> 4];
      *p++ = tbl[(unsigned char)(*q) & 15];
    }
    q++;
  }
  *p = '\0';
  return buf;
}

char* decode_qs(const char* q)
{
  char *buf, *p;
  int v;
  p = buf = (char*)malloc(sizeof(char) * strlen(q));
  while (*q) {
    if ((*q == '%')
     && (isxdigit(q[1]) && isxdigit(q[2]))) {
      sscanf(&(q[1]), "%02x", &v);
      if (v != 0x0d) {
	*p++ = (char)v;
      }
      q = &(q[3]);
    } else if (*q == '+') {
      *p++ = ' ';
      q++;
    } else if (*q == '&') {
      *p++ = '\n';
      q++;
    } else {
      *p++ = *q++;
    }
  }
  *p = '\0';
  return buf;
}

// compile expression
int split_qs(query_string** ms, const char *qs)
{
  char *ls, *rs;
  int  c, i, j;
  char *lbuf, *rbuf;

  if (!ms || !qs) {
    return -1;
  }

  *ms = (query_string*)malloc(0);
  lbuf = (char*)malloc(sizeof(char) * strlen(qs));
  rbuf = (char*)malloc(sizeof(char) * strlen(qs));
  c = 0;
  for (;;) {
    *ms = (query_string*)realloc(*ms, sizeof(query_string) * (c+1));
    while (*qs == '&') qs++;
    for (i = 0; *qs && (*qs != '='); i++, qs++) {
      lbuf[i] = *qs;
    }
    lbuf[i] = '\0';
    if (*qs == '\0') break;
    qs++; // skip '='

    for (j = 0; *qs && (*qs != '&'); j++, qs++) {
      rbuf[j] = *qs;
    }
    rbuf[j] = '\0';
    if (*qs) qs++; // skip '&'

    (*ms)[c].name = decode_qs(lbuf);
    (*ms)[c].value = decode_qs(rbuf);
    c++;
  }
  free(rbuf);
  free(lbuf);
  return c;
}

#if 0
main()
{
  char* test = "c=v&p=Welcome";
  query_string* map;
  int n;

  n = split_qs(&map, test);
  printf("%d: (%s,%s) (%s,%s)\n", n, map[0].name, map[0].value, 
	 map[1].name, map[1].value);
  return 0;
}
#endif
