/*
 * Pocket Wiki: C version
 * (c) 2001 TAKANO Ryousei
 * $Id: pwiki-c.l,v 1.12 2001/09/01 15:35:59 takano Exp $
 */
%x UL_LIST PREFORMAT ESCAPE
%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <dirent.h>
#include <regex.h>
#include "cgi.h"

#define HEAD_SECTION (2)
#define HEAD_SUBSECTION (3)
#define HEAD_SUBSUBSECTION (4)

#define N_BUF (512)
#define N_POS_FORM (5)

typedef struct {
  char* name;
  time_t mtime;
} page_list;

const char* DATA_DIR = "text/";
const char* INTER_WIKI_NAME = "InterWikiName";
const char* ALIAS_WIKI_NAME = "AliasWikiName";

const char* DEF_CONTENT = "Content-Type: text/html; charset=x-euc-jp\n";
const char* DEF_HEADER = "<html>
<head>
<meta http-equiv=\"Content-Type\" content=\"text/html;charset=x-euc-jp\">
<link rel=\"stylesheet\" type=\"text/css\" href=\"pwiki.css\">
<title>Pocket Wiki: %s</title>
</head>
<body>
<div align=\"right\">
 <a href=\"pwiki.cgi?c=v&p=%s\">Up</a>
|<a href=\"pwiki.cgi?c=e&p=%s\">Edit</a>
|<a href=\"pwiki.cgi?c=a\">AllPages</a>
|<a href=\"pwiki.cgi?c=s\">Search</a>
</div>
<h1>%s</h1>
";

const char* DEF_FORM_BEGIN = "<form action=\"pwiki.cgi\" method=\"post\">
<input type=\"hidden\" name=\"c\" value=\"W\">
<textarea name=\"text\" wrap=\"virtual\" cols=\"80\" rows=\"20\">";

const char* DEF_FORM_END = "</textarea><br>
<input type=\"hidden\" name=\"p\" value=\"%s\">
<input type=\"submit\" value=\"Save\">
</form>
";

const char* DEF_SEARCH_FORM = "<form action=\"pwiki.cgi\" method=\"post\">
<input type=\"hidden\" name=\"c\" value=\"R\">
Search Word: <input type=\"text\" name=\"text\">
<input type=\"submit\" value=\"Search\">
</form>
";

const char* DEF_FOOTER = "<div align=\"right\">
 <a href=\"pwiki.cgi?c=v&p=%s\">Up</a>
|<a href=\"pwiki.cgi?c=e&p=%s\">Edit</a>
|<a href=\"pwiki.cgi?c=a\">AllPages</a>
|<a href=\"pwiki.cgi?c=s\">Search</a>
</div>
</body>
</html>
";

#define MAX_STATES 32
int last_state[MAX_STATES] = {INITIAL};
int state_count = 0;
int list_level = 0;
const char* current_page;

typedef struct {
  int len;
  char*** list;
} indirect_wikis;

indirect_wikis inter_wikis, alias_wikis;

/* operate indirect wiki page */
#define READ_INTER_WIKIS() (read_indirect_wikis(INTER_WIKI_NAME, &inter_wikis))
#define READ_ALIAS_WIKIS() (read_indirect_wikis(ALIAS_WIKI_NAME, &alias_wikis))
#define IS_INTER_WIKIS(a, b) (is_indirect_wikis((a), (b), &inter_wikis))
#define IS_ALIAS_WIKIS(a, b) (is_indirect_wikis((a), (b), &alias_wikis))
#define IS_INDIRECT_WIKIS(a, b) (IS_INTER_WIKIS((a), (b)) || \
                                 IS_ALIAS_WIKIS((a), (b)))

void read_indirect_wikis(const char* name, indirect_wikis* wikis)
{
  int flen;
  char* fname;
  char buf[N_BUF];
  FILE* fp;

  flen = strlen(DATA_DIR) + strlen(name);
  fname = (char*)malloc(sizeof(char) * (flen+1));
  sprintf(fname, "%s%s", DATA_DIR, name);
  
  fp = fopen(fname, "r");
  if (fp == NULL) {
    free(fname);
    return;
  }

  wikis->len = 0;
  wikis->list = NULL;

  while (fgets(buf, N_BUF, fp)) {
    char *src, *dst, *tmp;

    buf[strlen(buf)-1] = '\0';
    tmp = strdup(buf);
    src = tmp;
    dst = strchr(tmp, ' ');
    if (dst == NULL) {
      free(tmp);
      continue;
    }
    *dst++ = '\0';

    wikis->list = (char***)realloc(wikis->list,
                                   (sizeof(char**) * (wikis->len+1)));
    wikis->list[wikis->len] = (char**)malloc(sizeof(char*) * 2);
    wikis->list[wikis->len][0] = src;
    wikis->list[wikis->len][1] = dst;
    wikis->len++;
  }
  free(fname);
  fclose(fp);
}

void init_indirect_wikis()
{
  READ_INTER_WIKIS();
  READ_ALIAS_WIKIS();
}

int is_indirect_wikis(const char* title, const char** atitle, indirect_wikis* wikis)
{
  int i;
  int ret = 0;
  char *iwiki, *name, *tmp;

  tmp = strdup(title);
  iwiki = tmp;
  name = strchr(tmp, ':');
  if (name == NULL) {
    iwiki = tmp;
  } else {
    *name = '\0';
  }

  for (i = 0; i < wikis->len; i++) {
    if (!strcmp(wikis->list[i][1], iwiki)) {
      ret = 1;
      *atitle = wikis->list[i][0];
      break;
    }
  }
  free(tmp);
  return ret;
}

void link_page(const char* name)
{
  int flen;
  char* fname;
  FILE* fp;
  char* ename;
  const char* alias;

  ename = encode_qs(name);
  if (IS_INDIRECT_WIKIS(name, &alias)) {
    printf("<a href=\"pwiki.cgi?c=v&p=%s\">%s</a>", ename, name);
    free(ename);
    return;
  }

  flen = strlen(DATA_DIR) + strlen(ename);
  fname = (char*)malloc(sizeof(char) * (flen+1));
  sprintf(fname, "%s%s", DATA_DIR, ename);

  fp = fopen(fname, "r");
  if (fp) {
    printf("<a href=\"pwiki.cgi?c=v&p=%s\">%s</a>", ename, name);
    fclose(fp);
  } else {
    printf("%s<a href=\"pwiki.cgi?c=e&p=%s\">?</a>", name, ename);
  }
  free(fname);
  free(ename);
}

void end_list(void)
{
  int i;

  for (i = list_level; i > 0; i--) {
    printf("</ul>\n");
  }
}

int sort_recently(const void* l1, const void* l2)
{
  return ((const page_list*)l1)->mtime < ((const page_list*)l2)->mtime;
}

int get_page_list(page_list** pages)
{
  DIR* dp;
  struct dirent* dent;
  struct stat sbuf;
  char fname[512]; // ...
  int nlist;
  page_list* tmp = NULL;

  dp = opendir(DATA_DIR);
  if (!dp) {
    return -1;
  }

  nlist = 0;
  strcpy(fname, DATA_DIR);
  while (dent = readdir(dp)) {
    if (dent->d_name[0] == '.') continue;

    strcpy(fname + strlen(DATA_DIR), dent->d_name);
    stat(fname, &sbuf);

    tmp = (page_list*)realloc(tmp, (nlist + 1) * sizeof(page_list));
    tmp[nlist].name = strdup(dent->d_name);
    tmp[nlist].mtime = sbuf.st_mtime;
    nlist++;
  }
  closedir(dp);

  if (nlist > 0) {
    qsort(tmp, nlist, sizeof(page_list), sort_recently);
  }
  *pages = tmp;
  return nlist;
}

void headline(int level, const char* title)
{
  printf("<h%d>%s</h%d>\n", level, title, level);
}
%}

%%
<*>"<"  { printf("&lt;"); }
<*>">"  { printf("&gt;"); }
<*>"&"  { printf("&amp;"); }
<*>"\"" { printf("&quot;"); }

<*>(http|https|ftp|mailto):[^ \n]* {
  printf("<a href=\"%s\">%s</a>", yytext, yytext);
}

<INITIAL,UL_LIST,PREFORMAT>[A-Z][a-z]+([A-Z][a-z]+)+ {
  link_page(yytext);
}

<INITIAL,UL_LIST,PREFORMAT>\[\[[^\*\/\{\}\'][^[:space:]]+\]\] {
  char* name;

  if (*(yytext + 2) == ':') {
    name = (char*)malloc(sizeof(char) * (strlen(current_page) + yyleng));
    strcpy(name, current_page);
  } else {
    name = (char*)malloc(sizeof(char) * yyleng);
    name[0] = '\0';
  }
  strncat(name, yytext+2, yyleng-4);
  link_page(name);
  free(name);
}

<INITIAL,UL_LIST,PREFORMAT>\[\[\* {
  printf("<strong>");
}
<INITIAL,UL_LIST,PREFORMAT>\*\]\] {
  printf("</strong>");
}

<INITIAL,UL_LIST,PREFORMAT>\[\[\/ {
  printf("<s>");
}
<INITIAL,UL_LIST,PREFORMAT>\/\]\] {
  printf("</s>");
}

<INITIAL,UL_LIST>\[\[\{ {
  printf("<pre>");
  last_state[++state_count] = PREFORMAT;
  BEGIN(PREFORMAT);
}
<PREFORMAT>\}\]\] { printf("</pre>"); BEGIN(last_state[--state_count]); }
<PREFORMAT><<EOF>> { printf("</pre>"); BEGIN(last_state[--state_count]); }

<INITIAL,UL_LIST>\[\[\' {
  last_state[++state_count] = ESCAPE;
  BEGIN(ESCAPE);
}
<ESCAPE>\'\]\] { BEGIN(last_state[--state_count]); }

<INITIAL>^"= "[^\n]* {
  headline(HEAD_SECTION, yytext+2);
}
<INITIAL>^"== "[^\n]* {
  headline(HEAD_SUBSECTION, yytext+3);
}
<INITIAL>^"=== "[^\n]* {
  headline(HEAD_SUBSUBSECTION, yytext+4);
}
<INITIAL>^----\n { printf("<hr>"); }
<INITIAL>[\n] { printf("<br>\n"); }
<INITIAL>^\$date {
  time_t now;

  time(&now);
  printf("%s", ctime(&now));
}
<INITIAL>^"  \* " {
  list_level = 1;
  printf("<ul>\n");
  printf("<li> ");
  last_state[++state_count] = UL_LIST;
  BEGIN(UL_LIST);
}
<UL_LIST>^"  "+"\* " {
  char* p;
  int nest;

  p = strchr(yytext, '*');
  nest = (p - yytext) / 2;
  if (nest > list_level) {
    list_level++;
    printf("<ul>\n");
  } else if (nest < list_level) {
    list_level--;
    printf("</ul>\n");
  }
  printf("<li> ");
}
<UL_LIST>^\n { end_list(); BEGIN(last_state[--state_count]); }
<UL_LIST><<EOF>> { end_list(); BEGIN(last_state[--state_count]); }
%%

/* get super page name */
char* get_super(char* name)
{
  char* p;
  char* buf;

  buf = strdup(name);
  p = strrchr(buf, ':');
  if (p == NULL) {
    return NULL;
  }
  *p = '\0';
  return buf;
}

/* view all Wiki pages */
int do_view_all()
{
  int nlist;
  int i;
  page_list* lists;

  nlist = get_page_list(&lists);
  if (nlist < 0) {
    fprintf(stderr, "PocketWiki: Can't open text dir!\n");
    return -1;
  }

  printf(DEF_CONTENT);
  printf("w3m-control: NEXT_LINK\n\n");
  printf(DEF_HEADER, "AllPages", "FrontPage", "AllPages", "AllPages");
  printf("<hr>\n<ul>\n");
  for (i = 0; i < nlist; i++) {
    char* dname;
    if (lists[i].name[0] == '.') continue;
    dname = decode_qs(lists[i].name);
    printf(" <li> <a href=\"pwiki.cgi?c=v&p=%s\">%s</a> %s\n", 
           lists[i].name, dname, ctime(&lists[i].mtime));
    free(dname);
  }
  printf("</ul>\n<hr>\n");
  printf(DEF_FOOTER, "FrontPage", "AllPages");
}

/* view Wiki page */
int do_view(const char* title)
{
  int flen;
  char* fname;
  struct stat sbuf;
  char* etitle;
  const char* indirect;
  char *sname, *up_page;

  current_page = title;
  if (IS_ALIAS_WIKIS(title, &indirect)) {
    current_page = indirect;
  }

  if (IS_INTER_WIKIS(current_page, &indirect)) {
    char* name;
    name = strchr(current_page, ':');
    name++;
    printf("Location: %s%s\n\n", indirect, name);
    return 0;
  }

  etitle = encode_qs(current_page);
  flen = strlen(DATA_DIR) + strlen(etitle);
  fname = (char*)malloc(sizeof(char) * (flen+1));
  sprintf(fname, "%s%s", DATA_DIR, etitle);
  yyin = fopen(fname, "r");
  if (!yyin) {
    do_edit(current_page);
    free(fname);
    return 0;
  }
  stat(fname, &sbuf);

  printf(DEF_CONTENT);
  printf("w3m-control: NEXT_LINK\n\n");
  sname = get_super((char*)current_page);
  up_page = sname ? encode_qs(sname) : "FrontPage";
  printf(DEF_HEADER, current_page, up_page, etitle, current_page);
  printf("<hr>\n");
  yylex();
  printf("<hr>\n(LastUpdate: %s)\n", ctime(&sbuf.st_mtime));
  printf(DEF_FOOTER, up_page, etitle);
  fclose(yyin);
  free(fname);
  free(sname);
}

/* edit Wiki page */
int do_edit(const char* title)
{
  int flen, len;
  char* fname;
  char* tmp;
  FILE* fp;
  char buf[N_BUF];
  char* etitle;
  char *sname, *up_page;
  int i;

  etitle = encode_qs(title);
  flen = strlen(DATA_DIR) + strlen(etitle);
  fname = (char*)malloc(sizeof(char) * (flen+1));
  sprintf(fname, "%s%s", DATA_DIR, etitle);
  fp = fopen(fname, "r");
  free(fname);

  len = strlen(title) + 5;
  tmp = (char*)malloc(sizeof(char) * (len+1));
  sprintf(tmp, "edit %s", title);

  printf(DEF_CONTENT);
  for (i = 0; i < N_POS_FORM; i++) { // w3m specific code!
    printf("w3m-control: NEXT_LINK\n");
  }
  printf("\n");
  sname = get_super(tmp);
  up_page = sname ? encode_qs(sname) : "FrontPage";
  printf(DEF_HEADER, tmp, up_page, etitle, title);
  printf(DEF_FORM_BEGIN);
  free(tmp);

  if (fp) {
    while (fgets(buf, N_BUF, fp)) {
      printf("%s", buf);
    }
    fclose(fp);
  }

  printf(DEF_FORM_END, title);
  printf(DEF_FOOTER, up_page, etitle);
  free(sname);
}

/* write Wiki page */
int do_write(const char* title, const char* text)
{
  int flen;
  char* fname;
  FILE* fp;
  char c;
  char* etitle;

  etitle = encode_qs(title);
  flen = strlen(DATA_DIR) + strlen(etitle);
  fname = (char*)malloc(sizeof(char) * (flen+1));
  sprintf(fname, "%s%s", DATA_DIR, etitle);
  fp = fopen(fname, "w");
  if (!fp) {
    do_view("FrontPage");
    return -1;
  }
  fwrite(text, sizeof(char) * strlen(text), 1, fp);
  fclose(fp);
  free(fname);

  do_view(title);
}

/* search Wiki page */
int do_search()
{
  int i;

  printf(DEF_CONTENT);
  for (i = 0; i < N_POS_FORM; i++) { // w3m specific code!
    printf("w3m-control: NEXT_LINK\n");
  }
  printf("\n");
  printf(DEF_HEADER, "Search", "FrontPage", "Search", "Search");
  printf("<hr>\n");
  printf(DEF_SEARCH_FORM);
  printf("<hr>\n");
  printf(DEF_FOOTER, "FrontPage", "Search");
}

int regexp_search(FILE* in, const char* text)
{
  char line[BUFSIZ];
  regex_t preg;
  regmatch_t pmatch[1];
  char* dtext;

  dtext = decode_qs(text);
  if (regcomp(&preg, dtext, REG_EXTENDED)) {
    return 0;
  }
  while (fgets(line, BUFSIZ, in) != NULL) {
    if (!regexec(&preg, line, 0, pmatch, 0)) {
      return 1;
    }
  }
  return 0;
}

int do_result(const char* text)
{
  DIR* dp;
  struct dirent* dent;
  int i;

  dp = opendir(DATA_DIR);
  if (!dp) {
    return -1;
  }
  chdir(DATA_DIR);

  printf(DEF_CONTENT);
  for (i = 0; i < N_POS_FORM; i++) { // w3m specific code!
    printf("w3m-control: NEXT_LINK\n");
  }
  printf("\n");
  printf(DEF_HEADER, "Search Results", "FrontPage", "Search Results", "Search Results");
  printf("<hr>\n<ul>\n");

  while (dent = readdir(dp)) {
    FILE* fp;
    if (dent->d_name[0] == '.') continue;
    fp = fopen(dent->d_name, "r");
    if (fp == NULL) continue;
    if (regexp_search(fp, text)) {
      char* dname;
      dname = decode_qs(dent->d_name);
      printf(" <li> <a href=\"pwiki.cgi?c=v&p=%s\">%s</a>\n", 
             dent->d_name, dname);
      free(dname);
    }
  }
  closedir(dp);

  printf("</ul>\n<hr>\n");
  printf(DEF_FOOTER, "FrontPage", "Search Results");
}

int yywrap()
{
  return 1;
}

#define CMD_UNKNOWN (-1)
#define CMD_VIEW    (0)
#define CMD_EDIT    (1)
#define CMD_WRITE   (2)
#define CMD_ALLPAGE (3)
#define CMD_SEARCH  (4)
#define CMD_RESULT_WORD   (5)
/* main */
int main(int argc, char** argv)
{
  query_string* qs;
  int n, i;
  int cmd = CMD_UNKNOWN;
  char* page = "FrontPage";
  char* text = NULL;

#if 0
  init_indirect_wikis();
  do_view("Omicron:Tiki");
#else
  n = split_qs(&qs, get_qs());
  init_indirect_wikis();
  for (i = 0; i < n; i++) {
//printf("[DEBUG: %s: %s(%d)]\n", qs[i].name, qs[i].value, strlen(qs[i].value));
    if (!strcmp(qs[i].name, "c")) {
      switch (qs[i].value[0]) {
      case 'a':
        cmd = CMD_ALLPAGE; break;
      case 'v':
        cmd = CMD_VIEW;    break;
      case 'e':
        cmd = CMD_EDIT;    break;
      case 'W':
        cmd = CMD_WRITE;   break;
      case 's':
        cmd = CMD_SEARCH;  break;
      case 'R':
        cmd = CMD_RESULT_WORD; break;
      default:
        cmd = CMD_UNKNOWN;
      }
    } else if (!strcmp(qs[i].name, "p")) {
      page = (char*)qs[i].value;
    } else if (!strcmp(qs[i].name, "text")) {
      text = (char*)qs[i].value;
    } else {
      cmd = CMD_UNKNOWN;
    }
  }
  switch (cmd) {
  case CMD_ALLPAGE:
    do_view_all();       break;
  case CMD_VIEW:
    do_view(page);       break;
  case CMD_EDIT:
    do_edit(page);       break;
  case CMD_WRITE:
    do_write(page, text); break;
  case CMD_SEARCH:
    do_search();         break;
  case CMD_RESULT_WORD:
    do_result(text);      break;
  default:
    do_view("FrontPage");
  }
#endif
}
